# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Not set tests."""

import pytest

from fourcipp.utils.not_set import NOT_SET, NotSet, check_if_set, pop_arguments


@pytest.mark.parametrize(
    "obj, expected_result", [("a", True), (2, True), ({2: 3}, True), (NOT_SET, False)]
)
def test_check_if_set(obj, expected_result):
    """Test check if set."""
    assert check_if_set(obj) == expected_result


def test_pop_arguments_without_default():
    """Assert pop arguments without default."""
    assert pop_arguments("key") == ("key",)


def test_pop_arguments_with_default():
    """Assert pop arguments with default."""
    assert pop_arguments("key", "default") == ("key", "default")


def test_expected():
    """Test expected."""
    a = NotSet(int)
    assert a.expected == int
